#include "jboci.h"

int ProcessQuery(HWND hWnd,char *query,char *schema,char *name,QueryProc do_function)
{
  char *char_ptr=NULL;
  int len=40,result;
  OCIBind *bndhp=NULL;
  OCIDefine *defhp=NULL;
  OCIStmt *stmtp=NULL;
  ub4 qrows;
  int nrows,x;
  char query_finished=0;
  ub4 total_qrows=0;
  int fetch_array_size=SCHEMA_ARRAY_SIZE;

  if (ChkOCIErrS(OCIHandleAlloc(g_oci->envhp, 
                               (dvoid**)&stmtp,
                               OCI_HTYPE_STMT, 
                               0, 
                               NULL)))
  {
    return -1;
  }
  if (ChkOCIErrS(OCIStmtPrepare(stmtp,
                               g_oci->errhp,
                               query,
                               strlen(query),
                               OCI_NTV_SYNTAX,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    stmtp=NULL;
    return -1;
  }
  if (ChkOCIErrS(OCIBindByName(stmtp,
                             &bndhp,
                             g_oci->errhp,
                             "schema_name",
                             strlen("schema_name"),
                             schema,
                             strlen(schema)+1,
                             SQLT_STR,
                             NULL,
                             NULL,
                             NULL,
                             0,
                             NULL,
                             OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    return -1;
  }
  if (ChkOCIErrS(OCIBindByName(stmtp,
                             &bndhp,
                             g_oci->errhp,
                             "object_name",
                             strlen("object_name"),
                             name,
                             strlen(name)+1,
                             SQLT_STR,
                             NULL,
                             NULL,
                             NULL,
                             0,
                             NULL,
                             OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    return -1;
  }
  char_ptr=(char*)malloc(SCHEMA_ARRAY_SIZE*len*sizeof(char));
  if (char_ptr==NULL)
  {
    Error(NULL,"Out of memory");
    return -2;
  }
  if (ChkOCIErrS(OCIDefineByPos(stmtp,
                               &defhp,
                               g_oci->errhp,
                               1,
                               char_ptr,
                               len,
                               SQLT_STR,
                               NULL,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    stmtp=NULL;
    free(char_ptr);
    return -1;
  }

  result=OCIStmtExecute(g_oci->svchp,
                        stmtp,
                        g_oci->errhp,
                        0,
                        0,
                        NULL,
                        NULL,
                        OCI_DEFAULT);
  if (result!=OCI_SUCCESS && result!=OCI_NO_DATA)
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    stmtp=NULL;
    free(char_ptr);
    return -1;
  }
  while (!query_finished)
  {
    result=OCIStmtFetch(stmtp,
                        g_oci->errhp,
                        fetch_array_size,
                        OCI_FETCH_NEXT,
                        OCI_DEFAULT);
    if (result!=OCI_SUCCESS && result!=OCI_NO_DATA && result!=OCI_SUCCESS_WITH_INFO)
    {
      if (ChkOCIErrI(result,1406))
        break;
    }
    if (ChkOCIErr(OCIAttrGet(stmtp,
                             OCI_HTYPE_STMT, 
                             &qrows,
                             NULL,
                             OCI_ATTR_ROW_COUNT, 
                             g_oci->errhp)))
    {
      break;
    } 
    nrows = qrows - total_qrows;
    if (nrows!=fetch_array_size)
      query_finished=1;

    for (x=0;x<(int)(qrows-total_qrows);x++)
    {
      if (do_function(hWnd,schema,&char_ptr[len*x]))
      {
        OCIHandleFree(stmtp,OCI_HTYPE_STMT);
        stmtp=NULL;
        free(char_ptr);
        return -1;
      }
    }
    total_qrows = qrows;
  }
  OCIHandleFree(stmtp,OCI_HTYPE_STMT);
  stmtp=NULL;
  free(char_ptr);
  return 0;
}

int GenerateSource(HWND hWnd,char *schema,char *name,char *type)
{
  OCIBind *bndhp=NULL;
  OCIDefine *defhp=NULL;
  OCIStmt *stmtp=NULL;
  char *columnstr="select text from all_source where owner=:p_schema and name=:p_name and type=:p_type order by line";
  char *text=NULL;
  ub2 *text_i=NULL;
  int result;
  ub4 qrows;
  int nrows,x;
  char query_finished=0;
  ub4 total_qrows=0;
  int fetch_array_size=SCHEMA_ARRAY_SIZE;
  char lastchar;

  if (ChkOCIErrS(OCIHandleAlloc(g_oci->envhp, 
                               (dvoid**)&stmtp,
                               OCI_HTYPE_STMT, 
                               0, 
                               NULL)))
  {
    return -1;
  }
  if (ChkOCIErrS(OCIStmtPrepare(stmtp,
                               g_oci->errhp,
                               columnstr,
                               strlen(columnstr),
                               OCI_NTV_SYNTAX,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    stmtp=NULL;
    return -1;
  }
  if (ChkOCIErrS(OCIBindByName(stmtp,
                             &bndhp,
                             g_oci->errhp,
                             "p_schema",
                             strlen("p_schema"),
                             schema,
                             strlen(schema)+1,
                             SQLT_STR,
                             NULL,
                             NULL,
                             NULL,
                             0,
                             NULL,
                             OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    return -1;
  }
  if (ChkOCIErrS(OCIBindByName(stmtp,
                             &bndhp,
                             g_oci->errhp,
                             "p_name",
                             strlen("p_name"),
                             name,
                             strlen(name)+1,
                             SQLT_STR,
                             NULL,
                             NULL,
                             NULL,
                             0,
                             NULL,
                             OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    return -1;
  }
  if (ChkOCIErrS(OCIBindByName(stmtp,
                             &bndhp,
                             g_oci->errhp,
                             "p_type",
                             strlen("p_type"),
                             type,
                             strlen(type)+1,
                             SQLT_STR,
                             NULL,
                             NULL,
                             NULL,
                             0,
                             NULL,
                             OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    return -1;
  }
  text=(char*)malloc(SCHEMA_ARRAY_SIZE*SOURCE_SIZE*sizeof(char));
  if (text==NULL)
  {
    Error(NULL,"Out of memory");
    return -2;
  }
  text_i=(ub2*)malloc(SCHEMA_ARRAY_SIZE*sizeof(ub2));
  if (text_i==NULL)
  {
    Error(NULL,"Out of memory");
    free(text);
    return -2;
  }
  if (ChkOCIErrS(OCIDefineByPos(stmtp,
                               &defhp,
                               g_oci->errhp,
                               1,
                               text,
                               SOURCE_SIZE,
                               SQLT_STR,
                               text_i,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    stmtp=NULL;
    free(text);
    free(text_i);
    return -1;
  }

  if (ChkOCIErrS(OCIStmtExecute(g_oci->svchp,
                        stmtp,
                        g_oci->errhp,
                        0,
                        0,
                        NULL,
                        NULL,
                        OCI_DEFAULT)))
  {
    OCIHandleFree(stmtp,OCI_HTYPE_STMT);
    stmtp=NULL;
    free(text);
    free(text_i);
    return -1;
  }
  query_finished=0;
  total_qrows=0;
  while (!query_finished)
  {
    result=OCIStmtFetch(stmtp,
                        g_oci->errhp,
                        fetch_array_size,
                        OCI_FETCH_NEXT,
                        OCI_DEFAULT);
    if (result!=OCI_SUCCESS && result!=OCI_NO_DATA && result!=OCI_SUCCESS_WITH_INFO)
    {
      if (ChkOCIErrI(result,1406))
        break;
    }
    if (ChkOCIErr(OCIAttrGet(stmtp,
                             OCI_HTYPE_STMT, 
                             &qrows,
                             NULL,
                             OCI_ATTR_ROW_COUNT, 
                             g_oci->errhp)))
    {
      break;
    } 
    nrows = qrows - total_qrows;
    if (nrows!=fetch_array_size)
      query_finished=1;

    for (x=0;x<(int)(qrows-total_qrows);x++)
    {
      if (text_i[x]!=(ub2)-1)
      {
        SchemaPrintf(hWnd,"%s",&text[4001*x]);
        lastchar=text[4001*x+strlen(&text[4001*x])-1];
      }
      else
      {
        SchemaPrintf(hWnd,"\n");
        lastchar=0;
      }
    }
    total_qrows = qrows;
  }
  if (lastchar!='\n')
  {
    SchemaPrintf(hWnd,"\n");
  }
  OCIHandleFree(stmtp,OCI_HTYPE_STMT);
  stmtp=NULL;
  free(text);
  free(text_i);
  return 0;
}