#include "jbsql.h"

#define TEXT_CLASS_NAME "TextClass"

DWORD ptid=0;

int GetMaxLineSize(TextWidget *text)
{
  int x,maxwidth=0;

  for (x=text->top;x<MIN(text->lines,text->top+text->height);x++)
  {
    maxwidth=MAX(maxwidth,(int)strlen(text->text[x]));
  }
  return maxwidth;
}

void AdjustScrollbars(HWND hwnd,TextWidget *text)
{
  SCROLLINFO si;
  int maxwidth;

  if (text->lines>text->height)
  {
    si.cbSize=sizeof(SCROLLINFO);
    si.fMask=SIF_POS|SIF_RANGE|SIF_PAGE;
    si.nMin=0;
    si.nMax=text->lines-1;
    si.nPos=text->top;
    si.nPage=text->height;
    SetScrollInfo(hwnd,SB_VERT,&si,TRUE);
    EnableScrollBar(hwnd,SB_VERT,ESB_ENABLE_BOTH);
  }
  else
  {
    EnableScrollBar(hwnd,SB_VERT,ESB_DISABLE_BOTH);
  }
  maxwidth=GetMaxLineSize(text);
  if (maxwidth>text->width)
  {
    si.cbSize=sizeof(SCROLLINFO);
    si.fMask=SIF_POS|SIF_RANGE|SIF_PAGE;
    si.nMin=0;
    si.nMax=maxwidth-1;
    si.nPos=text->left;
    si.nPage=text->width;
    SetScrollInfo(hwnd,SB_HORZ,&si,TRUE);
    EnableScrollBar(hwnd,SB_HORZ,ESB_ENABLE_BOTH);
  }
  else
  {
    EnableScrollBar(hwnd,SB_HORZ,ESB_DISABLE_BOTH);
  }
}

void UpdateCaretPosition(TextWidget *text,int scroll)
{
  int xpos,ypos,changed=0;

  if (text->active)
  {
    if (scroll && text->cur_pos>0 && text->cur_pos-1<text->left)
    {
      changed=1;
      text->left=MAX(text->cur_pos-10,0);
    }
    else if (scroll && text->cur_pos>0 && text->cur_pos-1>=text->left+text->width)
    {
      changed=1;
      text->left=text->cur_pos-text->width+10;
    }
    if (scroll && text->cur_line>0 && text->cur_line-1<text->top)
    {
      changed=1;
      text->top=text->cur_line-1;
    }
    else if (scroll && text->cur_line>0 && text->cur_line-1>=text->top+text->height)
    {
      changed=1;
      text->top=text->cur_line-text->height;
    }
    xpos=(text->cur_pos-text->left)*text->f_width+X_OFFSET;
    ypos=(text->cur_line-text->top-1)*text->f_height+Y_OFFSET+CARET_OFFSET;
    SetCaretPos(xpos,ypos);
    if (changed)
    {
      AdjustScrollbars(hwndTextBox,text);
      InvalidateRect(hwndTextBox, NULL, TRUE);
      UpdateWindow(hwndTextBox);
    }
  }
}


void GetRealSelection(TextWidget *text,
                      int *sel_left,
                      int *sel_top,
                      int *sel_right,
                      int *sel_bottom)
{
  if (text->sel_left<text->sel_right)
  {
    *sel_left=text->sel_left;
    *sel_right=text->sel_right;
  }
  else
  {
    *sel_left=text->sel_right;
    *sel_right=text->sel_left;
  }
  if (text->sel_top<text->sel_bottom)
  {
    *sel_top=text->sel_top;
    *sel_bottom=text->sel_bottom;
  }
  else
  {
    *sel_top=text->sel_bottom;
    *sel_bottom=text->sel_top;
  }
}

void RefreshTextWidget(TextWidget *text)
{
  HBRUSH hb;
  COLORREF c;
  RECT rct;
  int x,sel_left,sel_top,sel_bottom,sel_right;
  static char str[500],drawstr[500];

  SetTextColor(text->hdc, GetSysColor(COLOR_WINDOWTEXT));
  c = GetSysColor(COLOR_WINDOW);
  SetBkColor(text->hdc, c);
  hb = CreateSolidBrush(c);
  SelectObject(text->hdc,hb);
  SelectObject(text->hdc,text->hfont);

  for (x=text->top;x<MIN(text->top+text->height,text->lines);x++)
  {
    if ((int)strlen(text->text[x])>text->left)
    {
      strncpy(str,&text->text[x][text->left],MIN((int)strlen(&text->text[x][text->left]),text->width));
      str[MIN((int)strlen(&text->text[x][text->left]),text->width)]='\0';
    }
    else
      str[0]='\0';
    sprintf(drawstr,"%-*s",text->width,str);
    TextOut(text->hdc,X_OFFSET,Y_OFFSET+(x-text->top)*text->f_height,drawstr,text->width);
  }

  //Left
  rct.top = 0;
  rct.bottom = text->w_height;
  rct.left = 0;
  rct.right = X_OFFSET-1;
  FillRect(text->hdc, &rct, hb);

  //Top
  rct.top = 0;
  rct.bottom = Y_OFFSET-1;
  rct.left = 0;
  rct.right = text->w_width;
  FillRect(text->hdc, &rct, hb);

  //Right
  rct.top = 0;
  rct.bottom = text->w_height;
  rct.left = X_OFFSET+text->width*text->f_width;
  rct.right = text->w_width;
  FillRect(text->hdc, &rct, hb);

  //Bottom
  rct.top = Y_OFFSET+MIN(text->height,text->lines)*text->f_height;
  rct.bottom = text->w_height;
  rct.left = 0;
  rct.right = text->w_width;
  FillRect(text->hdc, &rct, hb);

  DeleteObject((HANDLE)hb);
  if (text->selection)
  {
    GetRealSelection(text,&sel_left,&sel_top,&sel_right,&sel_bottom);
    if (sel_left<=text->left+text->width && 
        sel_top<=text->top+text->height &&
        sel_bottom>=text->top &&
        sel_right>=text->left)
    {
      c=GetSysColor(COLOR_WINDOWTEXT);
      SetBkColor(text->hdc, c);
      hb = CreateSolidBrush(c);
      SelectObject(text->hdc,hb);
      SetROP2(text->hdc,R2_MERGEPENNOT);
      if (sel_left<text->left)
        sel_left=X_OFFSET;
      else
        sel_left=X_OFFSET+(sel_left-text->left)*text->f_width;
      if (sel_top<text->top)
        sel_top=Y_OFFSET;
      else
        sel_top=Y_OFFSET+(sel_top-text->top)*text->f_height;
      if (sel_right>text->left+text->width)
        sel_right=X_OFFSET+((text->w_width+1)*text->f_width);
      else
        sel_right=X_OFFSET+(sel_right-text->left+1)*text->f_width;
      if (sel_bottom>text->top+text->height)
        sel_bottom=Y_OFFSET+((text->height+1)*text->f_height);
      else
        sel_bottom=Y_OFFSET+(sel_bottom-text->top+1)*text->f_height;
      Rectangle(text->hdc,sel_left,sel_top,sel_right,sel_bottom);
      DeleteObject((HANDLE)hb);
    }
  }
  UpdateCaretPosition(text,0);
}

void ScrollTextWidgetV(HWND hwnd,TextWidget *text,int amount)
{
  int maxwidth=0;

  text->top+=amount;
  if (text->top>text->lines-text->height)
    text->top=text->lines-text->height;
  if (text->top<0)
    text->top=0;
  maxwidth=GetMaxLineSize(text);
  if (maxwidth<=text->width)
    text->left=0;
  AdjustScrollbars(hwnd,text);
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}

void ScrollTextWidgetH(HWND hwnd,TextWidget *text,int amount)
{
  int maxwidth=0;

  maxwidth=GetMaxLineSize(text);
  text->left+=amount;
  if (text->left>maxwidth-text->width)
    text->left=maxwidth-text->width;
  if (text->left<0)
    text->left=0;
  AdjustScrollbars(hwnd,text);
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}

void WPrintfOnly(char *fmt,...)
{
  va_list param;
  char *str;
  void *spoolfile;

  va_start(param,fmt);
  str=myvsprintf(fmt,param);
  va_end(param);

  spoolfile=g_text->spoolfile;
  g_text->spoolfile=NULL;
  WPrintf("%s",str);
  g_text->spoolfile=spoolfile;
}

void WPrintf(char *fmt,...)
{
  va_list param;
  char *str;
  int x,lastx=0,slen;
  TextWidget *text;

  va_start(param,fmt);
  str=myvsprintf(fmt,param);
  va_end(param);

  if (hwndTextBox==NULL)
  {
    free(str);
    return;
  }
  text = g_text;
  slen=strlen(str);
  if (text->spoolfile && tid)
    FPrintf("%s",str);
  for (x=0;x<(int)slen;x++)
  {
    switch (str[x])
    {
      case '\n':
        str[x]='\0';
        AppendLine(text,&str[lastx]);
        InsertLine(text);
        lastx=x+1;
        break;
      case '\r':
        ClearLine(text);
        lastx=x+1;
        break;
      case 8:
        DeleteCharacter(text);
        lastx=x+1;
        break;
      default:
        break;
    }
  }
  if (lastx<(int)slen)
    AppendLine(text,&str[lastx]);
/*  if (text->lines>text->height)
    text->top=text->lines-text->height;*/
  if (text->cur_line<text->top)
    text->top=text->cur_line;
  else if (text->cur_line>text->top+text->height)
    text->top=text->cur_line-text->height;
  if (text->cur_pos>text->left+text->width)
    text->left=text->cur_pos-text->width;
  if (text->noredraw==0)
  {
    AdjustScrollbars(hwndTextBox,text);
    InvalidateRect(hwndTextBox, NULL, TRUE);
    UpdateWindow(hwndTextBox);
  }
  free(str);
}

void WPrintfUP(char *str)
{
  if (hwndTextBox==NULL)
    return;
  if (g_text->spoolfile && tid)
    FPrintf("%s",str);
  AppendLine(g_text,str);
  if (g_text->cur_line<=g_text->top)
    g_text->top=g_text->cur_line;
/*  if (g_text->lines>g_text->height)
    g_text->top=g_text->lines-g_text->height;*/
  if (g_text->cur_pos>g_text->left+g_text->width)
    g_text->left=g_text->cur_pos-g_text->width;
  if (g_text->noredraw==0)
  {
    AdjustScrollbars(hwndTextBox,g_text);
    InvalidateRect(hwndTextBox, NULL, TRUE);
    UpdateWindow(hwndTextBox);
  }
}

void WPrintfNL(void)
{
  if (hwndTextBox==NULL)
    return;
  if (g_text->spoolfile && tid)
    FPrintf("\n");
  InsertLine(g_text);
  if (g_text->lines>g_text->height)
    g_text->top=g_text->lines-g_text->height;
  if (g_text->cur_pos>g_text->left+g_text->width)
    g_text->left=g_text->cur_pos-g_text->width;
  if (g_text->noredraw==0)
  {
    AdjustScrollbars(hwndTextBox,g_text);
    InvalidateRect(hwndTextBox, NULL, TRUE);
    UpdateWindow(hwndTextBox);
  }
}

void FPrintf(char *fmt,...)
{
  va_list param;
  char *str;
  int x,lastx=0,slen;
  TextWidget *text;
  DWORD bytes_written;

  va_start(param,fmt);
  str=myvsprintf(fmt,param);
  va_end(param);

  text = g_text;
  if (text->spoolfile==NULL)
  {
    free(str);
    return;
  }
  slen=strlen(str);
  for (x=0;x<(int)slen;x++)
  {
    switch (str[x])
    {
      case '\n':
        str[x]='\0';
        WriteFile(text->spoolfile,&str[lastx],x-lastx,&bytes_written,NULL);
        WriteFile(text->spoolfile,"\r\n",2,&bytes_written,NULL);
        lastx=x+1;
        break;
      case '\r':
        lastx=x+1;
        break;
      default:
        break;
    }
  }
  if (lastx<(int)slen)
    WriteFile(text->spoolfile,&str[lastx],x-lastx,&bytes_written,NULL);
  free(str);
}

void SetPrintfFont(HWND hwnd,HFONT hfont)
{
  TextWidget *text;
  TEXTMETRIC tm;
  HFONT      hfontOld;

  text = (TextWidget*)GetWindowLong(hwnd, 0);
  hfontOld = SelectObject(text->hdc, hfont);
  DeleteObject(hfontOld);
  text->hfont=hfont;
  GetTextMetrics(text->hdc, &tm);
  text->f_height = tm.tmHeight;
  text->f_width = tm.tmAveCharWidth;
}

void ResizeTextWidget(HWND hwnd)
{
  RECT  rc;
  TextWidget *text;

  GetClientRect(hwnd, &rc);
  text = (TextWidget*)GetWindowLong(hwnd, 0);
  text->w_width=rc.right-rc.left;
  text->w_height=rc.bottom-rc.top;
  text->width=(text->w_width-2*X_OFFSET)/text->f_width;
  text->height=(text->w_height-2*Y_OFFSET)/text->f_height;
  if (text->lines>text->height)
    text->top=text->lines-text->height;
  else
    text->top=0;
  if (text->cur_pos>text->width)
    text->left=text->cur_pos-text->width;
  else
    text->left=0;
  AdjustScrollbars(hwnd,text);
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}

char *GetTextSelection(TextWidget *text)
{
  char *str=NULL;
  int sel_left,sel_top,sel_right,sel_bottom,x,newsize;

  if (text->selection)
  {
    str=strdup("");
    GetRealSelection(text,&sel_left,&sel_top,&sel_right,&sel_bottom);
    for (x=sel_top;x<=MIN(sel_bottom,text->lines-1);x++)
    {
      if ((int)strlen(text->text[x])>sel_left)
      {
        newsize=(strlen(str)+MIN((int)strlen(&text->text[x][sel_left]),sel_right-sel_left+1)+1);
        str=realloc(str,newsize);
        strncpy(&str[strlen(str)],&text->text[x][sel_left],MIN((int)strlen(&text->text[x][sel_left]),sel_right-sel_left+1)+1);
        str[newsize-1]=0;
      }
      str=realloc(str,(strlen(str)+2));
      if (x<sel_bottom)
        strcat(str,"\n");
    }
  }
  return str;
}

void CopyToClipboard(TextWidget *text)
{
  char *str,*clipstr,*newclipstr,*lockedclipstr;
  int slen,x,lastx=0;

  if (text->selection)
  {
    str=GetTextSelection(text);
    clipstr=malloc(1);
    clipstr[0]='\0';
    slen=strlen(str);
    for (x=0;x<(int)slen;x++)
    {
      switch (str[x])
      {
        case '\n':
          str[x]='\0';
          clipstr=realloc(clipstr,strlen(clipstr)+strlen(&str[lastx])+3);
          strcat(clipstr,&str[lastx]);
          strcat(clipstr,"\r\n");
          lastx=x+1;
          break;
        default:
          break;
      }
    }
    if (lastx<slen)
    {
      clipstr=realloc(clipstr,strlen(clipstr)+strlen(&str[lastx])+1);
      strcat(clipstr,&str[lastx]);
    }
    if (clipstr==NULL)
      DisplayLastError(NULL);
    if (OpenClipboard(hwndTextBox))
    {
      EmptyClipboard();
      newclipstr=GlobalAlloc(GMEM_DDESHARE,strlen(clipstr)+1);
      if (newclipstr!=NULL)
      {
        lockedclipstr=GlobalLock(newclipstr);
        if (lockedclipstr!=NULL)
        {
          strcpy(lockedclipstr,clipstr);
          GlobalUnlock(newclipstr);
          if (SetClipboardData(CF_TEXT, newclipstr)==NULL)
            DisplayLastError(NULL);
        }
      }
      CloseClipboard();
    }
    free(clipstr);
    free(str);
  }
}    

void InsertPasteBuffer(TextWidget *text,char *str)
{
  int slen,x,lastx=0;

  slen=strlen(str);
  for (x=0;x<(int)slen;x++)
  {
    switch (str[x])
    {
      case '\r':
        str[x]='\0';
        if (strchr(&str[lastx],'\t'))
          InsertLineBuffer(text,"%s",&str[lastx]);
        else
          InsertUnparsedLineBuffer(text,&str[lastx]);
        InsertNewLineBuffer(text);
        while (tid!=0)
          Sleep(10);

        lastx=x+2;
        x++;
        break;
      case '\n':
        str[x]='\0';
        if (strchr(&str[lastx],'\t'))
          InsertLineBuffer(text,"%s",&str[lastx]);
        else
          InsertUnparsedLineBuffer(text,&str[lastx]);
        InsertNewLineBuffer(text);
        while (tid!=0)
          Sleep(10);

        lastx=x+1;
        break;
      default:
        break;
    }
  }
  if (lastx<slen)
  {
    if (strchr(&str[lastx],'\t'))
      InsertLineBuffer(text,"%s",&str[lastx]);
    else
      InsertUnparsedLineBuffer(text,&str[lastx]);
  }
}

int PasteFromClipboard(TextWidget *text)
{
  char *str;
  int format=0;

  if (OpenClipboard(hwndTextBox))
  {
    for (;;)
    {
      format=EnumClipboardFormats(format);
      if (format==0)
      {
        format=CF_TEXT;
        break;
      }
      else if (format==CF_OEMTEXT)
        break;
    }
    str=GetClipboardData(format);
    CloseClipboard();
    if (str!=NULL)
    {
      str=strdup(str);
      InsertPasteBuffer(text,str);
      free(str);
    }
    else
      DisplayLastError(NULL);
  }
  ptid=0;
  return 0;
}

void QuickPaste(TextWidget *text)
{
  char *str;

  str=GetTextSelection(text);
  InsertPasteBuffer(text,str);
  free(str);
  ptid=0;
}

VOID CALLBACK LeftSelectProc(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime)
{
  TextWidget *text;

  text=(TextWidget*)GetWindowLong(hwnd, 0);
  if (text->sel_right>0)
    text->sel_right--;
  if (text->sel_right<text->left)
    text->left=text->sel_right;
  AdjustScrollbars(hwnd,text);
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}

VOID CALLBACK RightSelectProc(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime)
{
  TextWidget *text;
  int maxwidth;

  text=(TextWidget*)GetWindowLong(hwnd, 0);
  maxwidth=GetMaxLineSize(text);
  if (text->sel_right<maxwidth)
    text->sel_right++;
  if (text->sel_right-text->width>text->left)
    text->left=text->sel_right-text->width;
  AdjustScrollbars(hwnd,text);
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}

VOID CALLBACK TopSelectProc(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime)
{
  TextWidget *text;

  text=(TextWidget*)GetWindowLong(hwnd, 0);
  if (text->sel_bottom>0)
    text->sel_bottom--;
  if (text->sel_bottom<text->top)
    text->top=text->sel_bottom;
  AdjustScrollbars(hwnd,text);
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}

VOID CALLBACK BottomSelectProc(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime)
{
  TextWidget *text;

  text=(TextWidget*)GetWindowLong(hwnd, 0);
  if (text->sel_bottom<text->lines)
    text->sel_bottom++;
  if (text->sel_bottom-text->height>text->top)
    text->top=text->sel_bottom-text->height;
  AdjustScrollbars(hwnd,text);
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}

void SetCaretPositionToMouse(TextWidget *text,int xpos,int ypos)
{
  int left,line,x,curline=0,found=0,curpos=0,startpos=0;

  left=(text->left+(xpos-X_OFFSET)/text->f_width)-strlen(STR_OPTION(PROMPT))-1;
  line=text->lines-(text->top+(ypos-Y_OFFSET)/text->f_height)-1;
  if (line==0)
  {
    found=1;
    curpos=strlen(command);
  }
  else if (line>0)
  {
    for (x=strlen(command)-1;x>=0;x--)
    {
      if (command[x]=='\n')
      {
        curline++;
        if (curline==line)
        {
          found=1;
          curpos=x;
          break;
        }
      }
    }
  }
  if (found)
  {
    for (x=curpos-1;x>=0;x--)
    {
      if (command[x]=='\n')
      {
        startpos=x+1;
        break;
      }
    }
    for (x=startpos;x<=(int)strlen(command);x++)
    {
      if (left<0 || command[x]=='\n' || command[x]=='\0' || x-startpos==left)
      {
        linepos=x;
        text->cur_pos=x-startpos+strlen(STR_OPTION(PROMPT))+1;
        text->cur_line=text->top+(ypos-Y_OFFSET)/text->f_height+1;
        UpdateCaretPosition(text,0);
        break;
      }
    }
  }
}

LRESULT CALLBACK TextWndProc(HWND   hwnd,
														 UINT   uiMessage,
														 WPARAM wParam,
														 LPARAM lParam)
{
  static int nokey=0;
  TextWidget *text;

  text=(TextWidget*)GetWindowLong(hwnd, 0);
  switch (uiMessage) 
	{
		case WM_CREATE:
			{
				HFONT ft;
        LPCREATESTRUCT csWindowLong = (LPCREATESTRUCT) lParam;

				text = (TextWidget*)csWindowLong->lpCreateParams;
				SetWindowLong(hwnd, 0, (long)text);
        text->hdc = GetDC(hwnd);
				ft=CreateFont(-MulDiv(10, GetDeviceCaps(text->hdc, LOGPIXELSY), 72),
						          0,
											0,
											0,
											FW_NORMAL,
											FALSE,
											FALSE,
											FALSE,
											ANSI_CHARSET,
											OUT_DEFAULT_PRECIS,
											CLIP_DEFAULT_PRECIS,
											DEFAULT_QUALITY,
											FIXED_PITCH|FF_MODERN,
											"Courier New");
				if (ft==NULL)
					DisplayLastError(NULL);
				SetPrintfFont(hwnd, ft);
        ReleaseDC(hwnd,text->hdc);
        EnableScrollBar(hwnd,SB_VERT,ESB_DISABLE_BOTH);
        EnableScrollBar(hwnd,SB_HORZ,ESB_DISABLE_BOTH);
			}
			break;

		case WM_SIZE:
      ResizeTextWidget(hwnd);
			break;

    case WM_SETFOCUS:
			SetFocus(hwnd);
      DestroyCaret();
      CreateCaret(hwnd,NULL,2,text->f_height-2*CARET_OFFSET);
      ShowCaret(hwnd);
      text->active=1;
      UpdateCaretPosition(text,0);
      break;

    case WM_KILLFOCUS:
      DestroyCaret();
      text->active=0;
      break;

		case WM_MOUSEACTIVATE:
		  SetFocus(hwnd);
			return MA_ACTIVATE;
    case WM_LBUTTONDOWN:
      if (text->selection)
      {
        text->selection=0;
        InvalidateRect(hwndTextBox, NULL, TRUE);
        UpdateWindow(hwndTextBox);
      }
      SetCapture(hwndTextBox);
      break;
    case WM_LBUTTONUP:
      ReleaseCapture();
      if (text->leftselect)
      {
        KillTimer(hwnd,text->leftselect);
        text->leftselect=0;
      }
      if (text->rightselect)
      {
        KillTimer(hwnd,text->rightselect);
        text->rightselect=0;
      }
      if (text->bottomselect)
      {
        KillTimer(hwnd,text->bottomselect);
        text->bottomselect=0;
      }
      if (text->topselect)
      {
        KillTimer(hwnd,text->topselect);
        text->topselect=0;
      }
      if (!text->selection && !tid)
        SetCaretPositionToMouse(text,LOWORD(lParam),HIWORD(lParam));
      break;
    case WM_RBUTTONDOWN:
      {
        if (text->selection)
        {
          if (ptid==0)
   			    CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)QuickPaste,text,0,&ptid);
        }
      }
      break;
    case WM_MOUSEMOVE:
      if (wParam&MK_LBUTTON)
      {
        if (!text->selection)
        {
          text->selection=1;
          text->sel_left=text->left+(LOWORD(lParam)-X_OFFSET)/text->f_width;
          text->sel_top=text->top+(HIWORD(lParam)-Y_OFFSET)/text->f_height;
        }
        if ((short)LOWORD(lParam)>=0)
        {
          if (LOWORD(lParam)<=text->w_width)
          {
            text->sel_right=text->left+(LOWORD(lParam)-X_OFFSET)/text->f_width;
            if (text->rightselect)
            {
              KillTimer(hwnd,text->rightselect);
              text->rightselect=0;
            }
          }
          else
          {
            if (!text->rightselect)
              text->rightselect=SetTimer(hwnd,RIGHT_TIMER,SCROLL_TIMEOUT,RightSelectProc);
          }
          if (text->leftselect)
          {
            KillTimer(hwnd,text->leftselect);
            text->leftselect=0;
          }
        }
        else
        {
          if (!text->leftselect)
            text->leftselect=SetTimer(hwnd,LEFT_TIMER,SCROLL_TIMEOUT,LeftSelectProc);
          if (text->rightselect)
          {
            KillTimer(hwnd,text->rightselect);
            text->rightselect=0;
          }
        }
        if ((short)HIWORD(lParam)>=0)
        {
          if (HIWORD(lParam)<text->w_height)
          {
            text->sel_bottom=text->top+(HIWORD(lParam)-Y_OFFSET)/text->f_height;
            if (text->bottomselect)
            {
              KillTimer(hwnd,text->bottomselect);
              text->bottomselect=0;
            }
          }
          else
          {
            if (!text->bottomselect)
              text->bottomselect=SetTimer(hwnd,BOTTOM_TIMER,SCROLL_TIMEOUT,BottomSelectProc);
          }
          if (text->topselect)
          {
            KillTimer(hwnd,text->topselect);
            text->topselect=0;
          }
        }
        else
        {
          if (!text->topselect)
            text->topselect=SetTimer(hwnd,TOP_TIMER,SCROLL_TIMEOUT,TopSelectProc);
          if (text->bottomselect)
          {
            KillTimer(hwnd,text->bottomselect);
            text->bottomselect=0;
          }
        }
        InvalidateRect(hwndTextBox, NULL, TRUE);
        UpdateWindow(hwndTextBox);
      }
      break;
		case WM_CHAR:
      if (nokey)
      {
        nokey--;
        break;
      }
      if (!tid)
        InsertLineBuffer(text,"%c",(char)wParam);
      if (text->hintlist)
        ShowHint(text,command,linepos);
			break;

		case WM_VSCROLL:
      switch (LOWORD(wParam)) 
			{
				case SB_LINEDOWN:
          ScrollTextWidgetV(hwnd,text,1);
          break;
				case SB_LINEUP:
          ScrollTextWidgetV(hwnd,text,-1);
          break;
				case SB_PAGEDOWN:
          ScrollTextWidgetV(hwnd,text,text->height);
          break;
				case SB_PAGEUP:
          ScrollTextWidgetV(hwnd,text,-text->height);
          break;
				case SB_THUMBTRACK:
				case SB_THUMBPOSITION:
          ScrollTextWidgetV(hwnd,text,HIWORD(wParam)-text->top);
					break;
				case SB_ENDSCROLL:
					break;
      }
			break;

		case WM_HSCROLL:
      switch (LOWORD(wParam)) 
			{
				case SB_LINEDOWN:
          ScrollTextWidgetH(hwnd,text,1);
          break;
				case SB_LINEUP:
          ScrollTextWidgetH(hwnd,text,-1);
          break;
				case SB_PAGEDOWN:
          ScrollTextWidgetH(hwnd,text,text->width);
          break;
				case SB_PAGEUP:
          ScrollTextWidgetH(hwnd,text,-text->width);
          break;
				case SB_THUMBTRACK:
				case SB_THUMBPOSITION:
          ScrollTextWidgetH(hwnd,text,HIWORD(wParam)-text->left);
					break;
				case SB_ENDSCROLL:
					break;
      }
			break;

		case WM_PAINT:
      {
        PAINTSTRUCT ps;

        text->hdc=BeginPaint(hwnd,&ps);
        RefreshTextWidget(text);
        EndPaint(hwnd,&ps);
      }
			break;

		case WM_KEYDOWN:
      if (tid)
        break;
			switch (wParam) 
			{
				case VK_UP:
          {
            int sel;

            if (text->scanf)
              break;
            if (text->hintlist!=NULL)
            {
              sel=SendMessage(text->hintlist,LB_GETCURSEL,0,0);
              if (sel==LB_ERR || sel==0)
                break;
              else
                sel--;
              SendMessage(text->hintlist,LB_SETCURSEL,sel,0);
            }
            else
            {
              if (linepos==(int)strlen(command))
                PreviousStatement(text);
              else
                MoveCursorUp(text);
            }
          }
					break;
				case VK_DOWN:
          {
            int sel;

            if (text->scanf)
              break;
            if (text->hintlist!=NULL)
            {
              sel=SendMessage(text->hintlist,LB_GETCURSEL,0,0);
              if (sel==LB_ERR)
                sel=0;
              else
                sel++;
              SendMessage(text->hintlist,LB_SETCURSEL,sel,0);
            }
            else
            {
              if (linepos==(int)strlen(command))
                NextStatement(text);
              else
                MoveCursorDown(text);
            }
          }
					break;
				case VK_LEFT:
          if (HIWORD(GetKeyState(VK_CONTROL)))
            MoveCursorLeft(text,1);
          else
            MoveCursorLeft(text,0);
					break;
				case VK_RIGHT:
          if (HIWORD(GetKeyState(VK_CONTROL)))
            MoveCursorRight(text,1);
          else
            MoveCursorRight(text,0);
					break;
				case VK_HOME:
          MoveCursorHome(text,0);
					break;
				case VK_END:
          MoveCursorEnd(text,0);
					break;
				case VK_PRIOR:
          MoveCursorHome(text,1);
					break;
				case VK_NEXT:
          MoveCursorEnd(text,1);
					break;
				case VK_RETURN:
          if (text->hintlist!=NULL)
          {
            SelectFromHint(text,linepos);
            nokey=1;
          }
          else
          {
            int old_newline=INT_OPTION(ALLOW_NEWLINE);

            nokey=1;
            if (HIWORD(GetKeyState(VK_SHIFT)))
            {
              INT_OPTION(ALLOW_NEWLINE)=1;
            }
            if (!INT_OPTION(ALLOW_NEWLINE))
            {
              linepos=strlen(command);
              SyncTextPosition(text);
              UpdateCaretPosition(text,1);
            }
            InsertNewLineBuffer(text);
            INT_OPTION(ALLOW_NEWLINE)=old_newline;
          }
					break;
				case VK_BACK:
					break;
        case VK_DELETE:
          if (DeleteNextLineBufferCharacter(text))
            DeleteNextCharacter(text);
          break;
        case VK_TAB:
          if (INT_OPTION(AUTOCOMPLETE) && ShowHint(text,command,linepos))
            nokey=1;
          break;
        case VK_ESCAPE:
          if (text->hintlist!=NULL)
          {
            DestroyHintList(text);
          }
          nokey=1;
          break;
        case VK_INSERT:
          if (HIWORD(GetKeyState(VK_CONTROL)))
          {
            if (text->selection)
              CopyToClipboard(text); 
          }
          else if (HIWORD(GetKeyState(VK_SHIFT)) && !tid)
          {
            if (ptid==0)
   			      CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PasteFromClipboard,text,0,&ptid);
          }
          break;

			}
			break;
		default:
			return DefWindowProc(hwnd, uiMessage, wParam, lParam);
  }
  return 0L;
}


HWND CreateTextWidget(HINSTANCE hInst,HWND parent,TextWidget **rtext)
{
	HWND        hwnd;
	WNDCLASSEX  wcex;
	ATOM        aReturn;
	DWORD       dwStyle;
  TextWidget  *text;

	wcex.cbSize          = sizeof(WNDCLASSEX);
	wcex.style           = 0;
	wcex.lpfnWndProc     = (WNDPROC)TextWndProc;
	wcex.cbClsExtra      = 0;
	wcex.cbWndExtra      = sizeof(TextWidget*);
	wcex.hInstance       = hInst;
	wcex.hCursor         = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground   = NULL;
	wcex.lpszMenuName    = NULL;
	wcex.lpszClassName   = TEXT_CLASS_NAME;
	wcex.hIcon           = NULL;
	wcex.hIconSm         = NULL;

	aReturn = RegisterClassEx(&wcex);
  if (aReturn==0)
    DisplayLastError(NULL);

	dwStyle =   WS_CHILD   |
		          WS_VSCROLL |
              WS_HSCROLL;
	
  text=(TextWidget*)malloc(sizeof(TextWidget));
  memset(text,0,sizeof(TextWidget));
  text->text=(char**)malloc(sizeof(char*));
  text->text[0]=strdup("");
  text->lines=1;
  text->cur_line=1;
  text->termout=1;

  command=StrdupCommand("");
  GetDefaultOptions(text);
  SetPWD(text);

  hwnd = CreateWindowEx( WS_EX_CLIENTEDGE,          // ex style
												 TEXT_CLASS_NAME,           // class name - defined in commctrl.h
												 "",                        // dummy text
												 dwStyle,                   // style
												 0,                         // x position
												 0,                         // y position
												 0,                         // width
												 0,                         // height
												 parent,                    // parent
												 (HMENU)ID_EDIT,            // ID
												 hInst,                     // instance
												 text);                     // no extra data
	if(!hwnd)
	{
		DisplayLastError(NULL);
	  return NULL;
  }
  ShowWindow(hwnd, SHOW_OPENWINDOW);
  *rtext=text;
  return hwnd;
}
